/*-
 * Copyright (C)2007 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2007 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: block.c,v 1.8 2007/03/10 04:12:27 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sys/socket.h>
#include <sysexits.h>
#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "pgsadump.h"
#include "block.h"
#include "compat.h"

#include "time.cc"
#include "types.cc"

static int
givepage(PGsaDumpCtl *ctl, PageHeaderData *pp, CompatPageHeader *p) {
    int ex = EX_OK;
    p->pd_size     =  ((size_t)&((PageHeaderData*)0)->pd_linp[0]);
    p->pd_lsn      =  pp->pd_lsn;
    p->pd_sui      =  pp->pd_sui;
    p->pd_lower    =  pp->pd_lower;
    p->pd_upper    =  pp->pd_upper;
    p->pd_special  =  pp->pd_special;
    p->pd_linp     = &pp->pd_linp[0];
    p->maxoff      =  PageIsEmpty(p) ? 0 : ((int)(pp->pd_lower - p->pd_size) / (int)sizeof(ItemIdData));

    switch (ctl->pd_ver) {
     case 0:
        p->pd_pagesize = pp->pd_pagesize_version;
        p->pd_version  = 0;
        break;
     case 1:
        p->pd_pagesize = pp->pd_pagesize_version & 0xFF00;
        p->pd_version  = pp->pd_pagesize_version & 0x00FF;
        break;
     default:
        ex = EX_DATAERR;
    }
    return(ex);
}

static int
giveitem(PGsaDumpCtl *ctl, HeapTupleHeaderData *tt, CompatTupleHeader *t) {
    int ex = EX_OK;

    switch (ctl->pd_ver) {
     case 0:
        t->t_size                =  (size_t)&((HeapTupleHeaderData*)0)->vp0.t_bits[0];
        t->t_oid                 =  tt->vp0.t_oid;
        t->t_cmin                =  tt->vp0.t_cmin;
        t->t_cmax                =  tt->vp0.t_cmax;
        t->t_xmin                =  tt->vp0.t_xmin;
        t->t_xmax                =  tt->vp0.t_xmax;
        t->t_ctid.ip_blkid.bi_hi =  tt->vp0.t_ctid.ip_blkid.bi_hi;
        t->t_ctid.ip_blkid.bi_lo =  tt->vp0.t_ctid.ip_blkid.bi_lo;
        t->t_ctid.ip_posid       =  tt->vp0.t_ctid.ip_posid;
        t->t_natts               =  tt->vp0.t_natts;
        t->t_infomask            =  tt->vp0.t_infomask;
        t->t_hoff                =  tt->vp0.t_hoff;
        t->t_bits                = &(tt->vp0.t_bits[0]);
        break;
     case 1:
        t->t_size                =  (size_t)&((HeapTupleHeaderData*)0)->vp1.t_bits[0];
        t->t_oid                 =  (tt->vp1.t_infomask & HEAP_HASOID)
                                  ? *((Oid*)((char*)tt + tt->vp1.t_hoff - sizeof(Oid)))
                                  : InvalidOid;
        t->t_cmin                =  tt->vp1.t_field2.t_cmin;
        t->t_cmax                =  tt->vp1.t_field3.t_cmax;
        t->t_xmin                =  tt->vp1.t_xmin;
        t->t_xmax                =  (tt->vp1.t_infomask & HEAP_XMAX_IS_XMIN)
                                  ? tt->vp1.t_xmin
                                  : tt->vp1.t_field2.t_xmax;
        t->t_ctid.ip_blkid.bi_hi =  tt->vp1.t_ctid.ip_blkid.bi_hi;
        t->t_ctid.ip_blkid.bi_lo =  tt->vp1.t_ctid.ip_blkid.bi_lo;
        t->t_ctid.ip_posid       =  tt->vp1.t_ctid.ip_posid;
        t->t_natts               =  tt->vp1.t_natts;
        t->t_infomask            =  tt->vp1.t_infomask;
        t->t_hoff                =  tt->vp1.t_hoff;
        t->t_bits                = &(tt->vp1.t_bits[0]);
        break;
     default:
        ex = EX_DATAERR;
    }
    return(ex);
}

static char
cmpsym(void *l, void *r) {
    if  ((char*)l < (char*)r) return('<');
    if  ((char*)l > (char*)r) return('>');
    return('=');
}

int
PGsaDumpBlock(PGsaDumpCtl *ctl, u_int32_t pgn, void *pp) {
    CompatPageHeader p;
    HeapTupleHeaderData *tt;
    CompatTupleHeader t;
    OffsetNumber n;
    ItemIdData *i;
    int j;
    int c; /* current print string is not empty */
    int e = EX_OK, ex = EX_OK; /* error code */
    void *q;

    ifBLIN_QV4(ctl->flags)
        fprintf(stderr, "+PGsaDumpBlock =%04X(%04X) N%d\n", ctl->atrterm, ctl->atrmask, ctl->atrmax);
    if  (!pp) {
        errno = EFAULT;
        ifBLIN_QV1(ctl->flags) warnx("Illegal page address");
        ex = EX_SOFTWARE;
    } else {
        if  (givepage(ctl, pp, &p)) {
            errno = EINVAL;
            ifBLIN_QV1(ctl->flags) warnx("Illegal version %d", ctl->pd_ver);
            ex = EX_DATAERR;
            goto end;
        }
        if  (PageIsEmpty(&p)) {
            ifBLIN_QV3(ctl->flags) fprintf(stderr, "Page %u is empty\n", pgn);
            goto end;
        }
        if  (p.pd_version != ctl->pd_ver) {
            errno = EINVAL;
            ifBLIN_QV1(ctl->flags) warnx("PG_PAGE_LAYOUT_VERSION=%d", p.pd_version);
            if  (ctl->flags & PGsaDumpXVER) {
                ex = EX_DATAERR;
                goto end;
        }   }
        if  (p.pd_pagesize != ctl->blksz) {
            errno = EINVAL;
            ifBLIN_QV1(ctl->flags) warnx("PageSize=%d", p.pd_pagesize);
            if  (ctl->flags & PGsaDumpXBSZ) {
                ex = EX_DATAERR;
                goto end;
        }   }
        ifBLIN_QV4(ctl->flags) fprintf( stderr, "lower=%04X upper=%04X special=%04X Ntuples=%u\n"
                                      , p.pd_lower, p.pd_upper, p.pd_special, p.maxoff
                                      );
        for (n = FirstOffsetNumber; n <= p.maxoff; n++) {
            if  (!(ctl->flags & PGsaDumpXTUP) && ex) break;
            i = &(p.pd_linp[n - 1]);
            tt = PageGetItem(pp, i);
            if  (giveitem(ctl, tt, &t)) {
                errno = EINVAL;
                ifBLIN_QV1(ctl->flags) warnx("Illegal version %d", ctl->pd_ver);
                ex = EX_DATAERR;
                goto end;
            }
            ifBLIN_QV3(ctl->flags) fprintf( stderr, "p=%08X t=%08X i=%08X(%04X+%04X)\n"
                                          , (u_int32_t)pp, (u_int32_t)tt, (u_int32_t)i
                                          , ItemIdGetOffset(i), ItemIdGetLength(i)
                                          );
            if  (!((ctl->flags & PGsaDumpLNUS) || ItemIdIsUsed(i))) {
                ifBLIN_QV4(ctl->flags) fprintf(stderr, "Item (%u,%d) not used\n", pgn, n);
                continue;
            }
            if  (!((ctl->flags & PGsaDumpLDEL) || !ItemIdDeleted(i))) {
                ifBLIN_QV4(ctl->flags) fprintf(stderr, "Item (%u,%d) deleted\n", pgn, n);
                continue;
            }
            if  (  (((char*)tt)            < ((char*)pp + p.pd_upper))
                || (((char*)tt + t.t_size) > ((char*)pp + ctl->blksz))
                || (((char*)tt + t.t_hoff) > ((char*)pp + ctl->blksz))
                ) {
                errno = EINVAL;
                ifBLIN_QV2(ctl->flags) printf( "Tuple out of block\n");
                ifBLIN_QV4(ctl->flags)
                    fprintf( stderr, "%08X          %c %08X+%08X\n"
                                     "%08X+%08X %c %08X+%08X\n"
                                     "%08X+%08X %c %08X+%08X\n"
                           , (u_int32_t)tt
                           , cmpsym((char*)tt, (char*)pp + p.pd_upper)
                           , (u_int32_t)pp, p.pd_upper
                           , (u_int32_t)tt, t.t_size
                           , cmpsym((char*)tt + t.t_size, (char*)pp + ctl->blksz)
                           , (u_int32_t)pp, ctl->blksz
                           , (u_int32_t)tt, t.t_hoff
                           , cmpsym((char*)tt + t.t_hoff, (char*)pp + ctl->blksz)
                           , (u_int32_t)pp, ctl->blksz
                           );
                ex = EX_DATAERR;
                continue;
            }
            if  (!(  (ctl->flags & PGsaDumpETID)
                  || (  (pgn == ((t.t_ctid.ip_blkid.bi_hi << 16) |  t.t_ctid.ip_blkid.bi_lo))
                     && (n == t.t_ctid.ip_posid)
                ) )  ) {
                ifBLIN_QV4(ctl->flags)
                    fprintf( stderr, "CTID: (%u,%d)<>(%u,%d)\n"
                           , pgn, n
                           , (t.t_ctid.ip_blkid.bi_hi << 16) |  t.t_ctid.ip_blkid.bi_lo
                           , t.t_ctid.ip_posid
                           );
                continue;
            }
            if  (t.t_natts < ctl->atrmin) {
                ifBLIN_QV4(ctl->flags) fprintf(stderr, "Nattrs: %u < min\n", t.t_natts);
                continue;
            }
            if  (t.t_natts > ctl->atrmax) {
                ifBLIN_QV4(ctl->flags) fprintf(stderr, "Nattrs: %u > max\n", t.t_natts);
                continue;
            }
            if  ((t.t_infomask ^ ctl->atrterm) & ctl->atrmask) {
                ifBLIN_QV4(ctl->flags) fprintf( stderr, "InfoMask %04X\n"
                                              , (t.t_infomask ^ ctl->atrterm) & ctl->atrmask
                                              );
                continue;
            }
            c = 0;
            if  (ctl->flags & PGsaDumpRTID) {
                printf(c ? "\t(%u,%d)" : "(%u,%d)", pgn, n);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpNATR) {
                printf(c ? "\t%u" : "%u", t.t_natts);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpINFO) {
                printf(c ? "\t%04X" : "%04X", t.t_infomask);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpCMAX) {
                printf(c ? "\t%u" : "%u", t.t_cmax);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpXMAX) {
                printf(c ? "\t%u" : "%u", t.t_xmax);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpCMIN) {
                printf(c ? "\t%u" : "%u", t.t_cmin);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpXMIN) {
                printf(c ? "\t%u" : "%u", t.t_xmin);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpOID) {
                printf(c ? "\t%u" : "%u", t.t_oid);
                c = 1;
            }
            if  (ctl->flags & PGsaDumpCTID) {
                printf( c ? "\t(%u,%d)" : "(%u,%d)"
                      , (t.t_ctid.ip_blkid.bi_hi << 16) |  t.t_ctid.ip_blkid.bi_lo
                      , t.t_ctid.ip_posid
                      );
                c = 1;
            }
            if  (ctl->flags & PGsaDumpNULS) {
                if  (c) printf("\t");
                for (j = 0; j < BITMAPLEN(t.t_natts); j++) printf("%02X", t.t_bits[j]);
                c = 1;
            }
            q = (char*)tt + t.t_hoff; /* tuples begin */
            e = EX_OK;
            for (j = 0; j < ctl->atrs[0] && !e; j++) {
            ;   if  (!(ctl->flags & PGsaDumpXATR || j < t.t_natts)) break;
            ;   if  (c) printf("\t");
            ;   if  (  ((t.t_infomask & HEAP_HASNULL) && !(t.t_bits[j >> 3] & (1 << (j & 7))))
                    || !(j < t.t_natts)
                    ) {
                    printf("\\N");
            ;   } else {
                    size_t len;
                    u_int32_t type = ctl->atrs[j + 1] & 0x7FFF;
                    u_int32_t array = ctl->atrs[j + 1] & 0x8000;
                    u_int32_t al = PGsaDumpCollection[ctl->pd_ver][type].align;
            ;       if  (array || al > ctl->alignmax) al = ctl->alignmax;
            ;       al--;
            ;       ifBLIN_QV3(ctl->flags)
                        fprintf( stderr, "p=%08X t=%08X q=%08X a=%08X"
                               , (u_int32_t)pp, (u_int32_t)tt, (u_int32_t)q, al
                               );
            ;       q = (void*)(((u_int32_t)q + al) & ~al);
            ;       if  (array || (PGsaDumpCollection[ctl->pd_ver][type].len < 0))
                        len = *((u_int32_t*)q);
                      else len = PGsaDumpCollection[ctl->pd_ver][type].len;
            ;       ifBLIN_QV3(ctl->flags)
                        fprintf( stderr, " q=%08X N=%d L=%d%s\n"
                               , (u_int32_t)q, j + 1, len, array ? " A" : ""
                               );
            ;       if  (  (((char*)q + len) < ((char*)q))
                        || (((char*)q + len) > ((char*)pp + ctl->blksz))
                        || (((char*)q + len) > ((char*)tt + ItemIdGetLength(i)))
                        ) {
            ;           ifBLIN_QV2(ctl->flags) printf("Item out of tuple\n");
            ;           ifBLIN_QV4(ctl->flags)
                            fprintf( stderr, "%08X+%08X %c %08X\n"
                                             "%08X+%08X %c %08X+%08X\n"
                                             "%08X+%08X %c %08X+%08X\n"
                                   , (u_int32_t)q, len
                                   , cmpsym((char*)q + len, (char*)q)
                                   , (u_int32_t)q
                                   , (u_int32_t)q, len
                                   , cmpsym((char*)q + len, (char*)pp + ctl->blksz)
                                   , (u_int32_t)pp, ctl->blksz
                                   , (u_int32_t)q, len
                                   , cmpsym((char*)q + len, (char*)tt + ItemIdGetLength(i))
                                   , (u_int32_t)tt, ItemIdGetLength(i)
                                   );
            ;           e = EX_DATAERR;
            ;           break;
            ;       }
            ;       if  (array) {
                        int k, ndim, *dim;
                        void *d;
                        u_int64_t osize;

            ;           ndim = ((ArrayType*)q)->ndim;
            ;           if  (ndim > MAXDIM) {
            ;               ifBLIN_QV2(ctl->flags) fprintf(stderr, "%d- \n", ndim);
            ;               e = EX_DATAERR;
            ;               break;
            ;           }
            ;           dim = (int*)(((char*)q) + sizeof(ArrayType));
            ;           osize = 1;
            ;           for (k = 0; k < ndim; k++) {
            ;               osize *= dim[k];
            ;               if  (osize > len) {
            ;                   osize = 0;
            ;                   break;
            ;           }   }
            ;           if  (!osize) {
            ;               ifBLIN_QV2(ctl->flags) fprintf(stderr, " \n");
            ;               e = EX_DATAERR;
            ;               break;
            ;           }
            ;           d = (void*)(2 * sizeof(int) * ndim + sizeof(ArrayType) + (char *)q);
            ;           d = (void*)(((u_int32_t)d + ctl->alignmax - 1) & ~(ctl->alignmax - 1));
            ;           e = PGsaD_array(ctl, dim, ndim, d, type);
            ;       } else {
            ;           e = (*(PGsaDumpCollection[ctl->pd_ver][type].conv))(ctl, q, 0);
            ;       }
            ;       q = (char*)q + len;
            ;   }
            ;   c = 1;
            }
            if  (c) printf("\n");
            if  (!(ctl->flags & PGsaDumpXCOL) && e) break;
    }   }
end:
    ifBLIN_QV4(ctl->flags) fprintf(stderr, "-PGsaDumpBlock %d\n", ex ? ex : e);
    return(ex ? ex : e);
}
